from django import template
from django.core.urlresolvers import reverse, NoReverseMatch
from django.template.defaultfilters import stringfilter
from django.utils.safestring import mark_safe, SafeData
from django.utils.html import escape


from emoji import Emoji

register = template.Library()


@register.filter(name='emoji_replace', is_safe=True, needs_autoescape=True)
@stringfilter
def emoji_replace(value, autoescape=None):
    autoescape = autoescape and not isinstance(value, SafeData)
    if autoescape:
        value = escape(value)
    return mark_safe(Emoji.replace(value))


@register.filter(name='emoji_replace_unicode', is_safe=True,
                 needs_autoescape=True)
@stringfilter
def emoji_replace_unicode(value, autoescape=None):
    autoescape = autoescape and not isinstance(value, SafeData)
    if autoescape:
        value = escape(value)
    return mark_safe(Emoji.replace_unicode(value))


@register.filter(name='emoji_replace_html_entities',
                 is_safe=True, needs_autoescape=True)
@stringfilter
def emoji_replace_html_entities(value, autoescape=None):
    # Replaced before because it needs unescaped &
    value = Emoji.replace_html_entities(value)
    autoescape = autoescape and not isinstance(value, SafeData)
    if autoescape:
        value = escape(value)
    return mark_safe(value)


@register.simple_tag
def emoji_load():
    try:
        url = reverse('emoji:list.json')
    except NoReverseMatch:
        return ''
    else:
        return "Emoji.setDataUrl('{0}').load();".format(url)



import os
import re
import struct
from sys import version_info

from django.contrib.staticfiles.storage import staticfiles_storage

try:
    from ._unicode_characters import UNICODE_ALIAS
except ImportError as exc:
    UNICODE_ALIAS = {}


from . import settings

__all__ = ('Emoji',)


UNICODE_WIDE = True
try:
    unichr(0x0001f48b)
except ValueError:  # pragma: no cover
    import unicodedata

    UNICODE_WIDE = False
    UNICODE_SURROGATE_MIN = 55296  # U+D800
    UNICODE_SURROGATE_MAX = 57343  # U+DFFF

    def convert_unicode_surrogates(surrogate_pair):
        return unicodedata.normalize('NFKD', surrogate_pair)
except NameError:
    unichr = chr  # Python3 doesn't have unichr

PYTHON3 = False
if version_info[0] == 3:
    PYTHON3 = True
else:
    from _python2 import hex_to_unicode


class Emoji(object):
    """Test if an emoji exists in the library and returns the URL to it.
    Also can add emojis to a text if they match the pattern :emoticon:.

    Usage:
    >>> emoji = Emoji()
    >>> 'dog' in emoji
    True
    >>> 'doesntexistatall' in emoji
    False
    >>> emoji['dog']  # Uses staticfiles app internally
    '/static/emoji/img/dog.png'
    >>> emoji.replace("I am a :cat:.")
    'I am a <img src="/static/emoji/img/cat.png" alt="cat" class="emoji">.'

    This class is a singleton and if imported as following an instantiated
    version will be imported.
    >>> from emoji import Emoji
    >>> Emoji['dog']
    '/static/emoji/dog.png'

    """
    _static_path = 'emoji/img'
    _image_path = os.path.join(os.path.dirname(__file__),
                               'static', 'emoji', 'img')
    _instance = None
    _pattern = re.compile(r':([a-z0-9\+\-_]+):', re.I)
    _files = []
    _unicode_characters = UNICODE_ALIAS

    # This character acts as a modifier, if it's ever seen then remove
    # it because the modification is done when converting to an image
    # anyway.
    _unicode_modifiers = (u'\ufe0e', u'\ufe0f')

    # HTML entities regexs
    _html_entities_integer_unicode_regex = re.compile(r'&#([0-9]+);')
    _html_entities_hex_unicode_regex = re.compile(r'&#x([0-9a-f]+);', re.I)

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super(Emoji, cls).__new__(cls, *args, **kwargs)

        return cls._instance

    def __init__(self):
        self.names()

    def __contains__(self, value):
        return value in self._files

    def keys(self):
        return self._files

    def __getitem__(self, item):
        if item in self._files:
            return self._static_url(item)

    def _static_url(self, name):
        return staticfiles_storage.url(
            '{0}/{1}.png'.format(self._static_path, name)
        )

    def _image_string(self, filename, alt=None):
        title = ' '.join(filename.split('_'))

        return settings.EMOJI_IMG_TAG.format(
            self._static_url(filename),
            alt or title,
            title,
        )

    @classmethod
    def names(cls):
        """A list of all emoji names without file extension."""
        if not cls._files:
            for f in os.listdir(cls._image_path):
                if(not f.startswith('.') and
                   os.path.isfile(os.path.join(cls._image_path, f))):
                    cls._files.append(os.path.splitext(f)[0])

        return cls._files

    @classmethod
    def replace(cls, replacement_string):
        """Add in valid emojis in a string where a valid emoji is between ::"""
        e = cls()

        def _replace_emoji(match):
            val = match.group(1)
            if val in e:
                return e._image_string(match.group(1))
            else:
                return match.group(0)

        return e._pattern.sub(_replace_emoji, replacement_string)

    @classmethod
    def replace_unicode(cls, replacement_string):
        """This method will iterate over every character in
        ``replacement_string`` and see if it mathces any of the
        unicode codepoints that we recognize. If it does then it will
        replace that codepoint with an image just like ``replace``.

        NOTE: This will only work with Python versions built with wide
        unicode caracter support. Python 3 should always work but
        Python 2 will have to tested before deploy.

        """
        e = cls()
        output = []
        surrogate_character = None

        if settings.EMOJI_REPLACE_HTML_ENTITIES:
            replacement_string = cls.replace_html_entities(replacement_string)

        for i, character in enumerate(replacement_string):
            if character in cls._unicode_modifiers:
                continue

            # Check whether this is the first character in a Unicode
            # surrogate pair when Python doesn't have wide Unicode
            # support.
            #
            # Is there any reason to do this even if Python got wide
            # support enabled?
            if(not UNICODE_WIDE and not surrogate_character and
               ord(character) >= UNICODE_SURROGATE_MIN and
               ord(character) <= UNICODE_SURROGATE_MAX):
                surrogate_character = character
                continue

            if surrogate_character:
                character = convert_unicode_surrogates(
                    surrogate_character + character
                )
                surrogate_character = None

            name = e.name_for(character)
            if name:
                if settings.EMOJI_ALT_AS_UNICODE:
                    character = e._image_string(name, alt=character)
                else:
                    character = e._image_string(name)

            output.append(character)

        return ''.join(output)

    @classmethod
    def name_for(cls, character):
        for modifier in cls._unicode_modifiers:
            character = character.replace(modifier, '')

        return cls._unicode_characters.get(character, False)

    @classmethod
    def replace_html_entities(cls, replacement_string):
        """Replaces HTML escaped unicode entities with their unicode
        equivalent. If the setting `EMOJI_REPLACE_HTML_ENTITIES` is
        `True` then this conversation will always be done in
        `replace_unicode` (default: True).

        """
        def _hex_to_unicode(hex_code):
            if PYTHON3:
                hex_code = '{0:0>8}'.format(hex_code)
                as_int = struct.unpack('>i', bytes.fromhex(hex_code))[0]
                return '{0:c}'.format(as_int)
            else:
                return hex_to_unicode(hex_code)

        def _replace_integer_entity(match):
            hex_val = hex(int(match.group(1)))

            return _hex_to_unicode(hex_val.replace('0x', ''))

        def _replace_hex_entity(match):
            return _hex_to_unicode(match.group(1))

        # replace integer code points, &#65;
        replacement_string = re.sub(
            cls._html_entities_integer_unicode_regex,
            _replace_integer_entity,
            replacement_string
        )
        # replace hex code points, &#x41;
        replacement_string = re.sub(
            cls._html_entities_hex_unicode_regex,
            _replace_hex_entity,
            replacement_string
        )

        return replacement_string



from django.conf import settings

# The exact tag that is being used for replacing images and the values
# being passed in through `str#format`.
EMOJI_IMG_TAG = getattr(settings, 'EMOJI_IMG_TAG', (
    u'<img src="{0}" alt="{1}" title="{2}" class="emoji">'
))

EMOJI_ALT_AS_UNICODE = getattr(settings, 'EMOJI_ALT_AS_UNICODE', True)
EMOJI_REPLACE_HTML_ENTITIES = getattr(settings, 'EMOJI_REPLACE_HTML_ENTITIES',
                                      True)



import json

from django.http import HttpResponse
from django.views.generic import View

from . import Emoji


class EmojiJSONListView(View):
    def get(self, request, *args, **kwargs):
        return HttpResponse(json.dumps(dict(Emoji)),
                            content_type='application/json')



from django.conf.urls import patterns, include, url
from django.views.generic import TemplateView

from test.views import (
    EmojiTestReplaceTagView,
    EmojiTestReplcaceUnicodeTagView,
    EmojiTestReplaceHtmlEntitiesView,
    EmojiTestXSSFix,
)


urlpatterns = patterns(
    '',
    url(r'', include('emoji.urls', app_name='emoji', namespace='emoji')),
    url(r'^all$', EmojiTestReplaceTagView.as_view(), name='emoji_test_list'),
    url(r'^replace-unicode/$', EmojiTestReplcaceUnicodeTagView.as_view(),
        name='emoji_replace_unicode_test'),
    url(r'^replace-html-entities/$',
        EmojiTestReplaceHtmlEntitiesView.as_view(),
        name='emoji_replace_html_entities_test'),
    url(r'^xss-fix/$', EmojiTestXSSFix.as_view(), name='xss_fix'),
    url(r'^emoji-include-test$',
        TemplateView.as_view(template_name='emoji_include_tag.html'),
        name='emoji_include_test'),
    url(r'^emoji-load-test$',
        TemplateView.as_view(template_name='emoji_load_tag.html'),
        name='emoji_load_test'),
)



# encoding: utf-8
import os
from setuptools import setup, find_packages

README = open(os.path.join(os.path.dirname(__file__), 'README.rst')).read()

os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='django-emoji',
    version='2.0.0',
    packages=find_packages(exclude=('test',)),
    include_package_data=True,
    license='BSD License',
    description='A simple django app to use emojis on your website',
    long_description=README,
    url='https://github.com/gaqzi/django-emoji/',
    author='Björn Andersson',
    author_email='ba@sanitarium.se',
    install_requires=[
        'django',
    ],
    tests_require=[
        'django_nose',
    ],
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
    ],
)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals
import json

from django.core.urlresolvers import reverse, NoReverseMatch
from django.test import TestCase

from . import Emoji as EmojiInstance
from .models import Emoji, UNICODE_WIDE

# Anyone know how to mock out so os.listdir only lists what I want
# instead of hitting the file system?
TOTAL_EMOJIS = 888
TOTAL_EMOJIS_UNICODE = 838


class EmojiTest(TestCase):
    def test_emoji_is_singleton(self):
        e = Emoji()
        e2 = Emoji()

        self.assertEqual(id(e), id(e2))

    def test_get_all_names(self):
        self.assertEqual(len(Emoji.names()), TOTAL_EMOJIS)

    def test_name_is_without_file_suffix(self):
        first_file = Emoji.names()[0]
        self.assertTrue('.png' not in first_file)

    def test_in_should_work_when_looking_for_emoji(self):
        emoji = Emoji()
        self.assertTrue('relaxed' in emoji)

    def test_replace_string_with_emojis(self):
        test_string = '''It's raining :cat:s and :dog:s.'''
        replaced_string = Emoji.replace(test_string)
        self.assertTrue('cat.png' in replaced_string)
        self.assertTrue('dog.png' in replaced_string)

    def test_replace_string_with_emojis_no_emoji_do_nothing(self):
        self.assertEqual(Emoji.replace('Hello.'), 'Hello.')

    def test_if_no_valid_emoji_return_unchanged(self):
        self.assertEqual(Emoji.replace(':ducklingdrinkingcoffee:'),
                         ':ducklingdrinkingcoffee:')

    def test_replace_string_should_use_staticfiles_app(self):
        self.assertEqual(
            Emoji.replace(':+1:'),
            '<img src="/static/emoji/img/%2B1.png" alt="+1" '
            'title="+1" class="emoji">'
        )

    def test_emojiinstance_from_app(self):
        self.assertTrue('dog' in EmojiInstance)

    def test_convert_instance_to_dict_and_act_like_dict(self):
        e = Emoji()
        as_dict = dict(e)

        self.assertEqual(len(as_dict), TOTAL_EMOJIS)
        self.assertEqual(as_dict['+1'], '/static/emoji/img/%2B1.png')
        with self.assertRaises(KeyError):
            as_dict['nonexistant']


class EmojiJSONListViewTest(TestCase):
    def test_should_return_json_list(self):
        res = self.client.get(reverse('emoji:list.json'))
        self.assertEqual(res.status_code, 200)

        items = json.loads(res.content.decode('utf8'))
        self.assertEqual(len(items), TOTAL_EMOJIS)
        self.assertEqual(items['+1'], '/static/emoji/img/%2B1.png')


class EmojiTemplateTagTest(TestCase):
    def test_emoji_replace_tag(self):
        try:
            res = self.client.get(reverse('emoji_test_list'), {'limit': 1})
        except NoReverseMatch:
            return

        self.assertEqual(res.status_code, 200)
        self.assertIn('<img src="/static/emoji/img/%2B1.png" '
                      'alt="+1" title="+1" class="emoji">', str(res.content))

    def test_emoji_replace_unicode_tag(self):
        try:
            res = self.client.get(reverse('emoji_replace_unicode_test'))
        except NoReverseMatch:
            return

        self.assertEqual(res.status_code, 200)
        self.assertIn(('<img src="/static/emoji/img/kiss.png" ' +
                       'alt="💋" title="kiss" ' +
                       'class="emoji">'), res.content.decode('utf-8'))

    def test_emoji_replace_html_entities(self):
        try:
            res = self.client.get(reverse('emoji_replace_html_entities_test'))
        except NoReverseMatch:
            return

        self.assertEqual(res.status_code, 200)
        self.assertEqual(res.content.decode('utf-8'),
                         '\n\U0001f48b\n\U0001f48b\n')

    def test_emoji_include_script(self):
        try:
            res = self.client.get(reverse('emoji_include_test'))
        except NoReverseMatch:
            return

        self.assertEqual(res.status_code, 200)
        body = res.content
        self.assertTrue('/static/emoji/js/emoji.js' in str(body), 'emoji.js')

    def test_emoji_load_tag(self):
        try:
            res = self.client.get(reverse('emoji_load_test'))
        except NoReverseMatch:
            pass

        self.assertEqual(res.status_code, 200)
        self.assertTrue(
            "Emoji.setDataUrl('/all.json').load();" in str(res.content)
        )

    def test_xss_replace_html_entities(self):
        try:
            res = self.client.get(reverse('xss_fix'))
        except NoReverseMatch:
            pass

        self.assertEqual(res.status_code, 200)
        correct_string = ('&lt;em&gt;\U0001f48b&lt;/em&gt;\n'
                          '&lt;em&gt;\U0001f48b&lt;/em&gt;')
        self.assertIn(correct_string, res.content.decode('utf-8'))

    def test_xss_replace_unicode(self):
        try:
            res = self.client.get(reverse('xss_fix'))
        except NoReverseMatch:
            pass

        self.assertEqual(res.status_code, 200)
        correct_string = ('&lt;em&gt;<img src="/static/emoji/img/shit.png" '
                          'alt="💩" title="shit" class="emoji">&lt;em&gt;')
        self.assertIn(correct_string, res.content.decode('utf-8'))

    def test_xss_replace_emoji(self):
        try:
            res = self.client.get(reverse('xss_fix'))
        except NoReverseMatch:
            pass

        self.assertEqual(res.status_code, 200)
        correct_string = ('&lt;em&gt;<img src="/static/emoji/img/dog.png" '
                          'alt="dog" title="dog" class="emoji">&lt;/em&gt;')
        self.assertIn(correct_string, res.content.decode('utf-8'))


class UnicodeTestBase(TestCase):
    # :kiss: filename 1f48b.png, chosen 'cause it prints properly in terminal
    UNICODE_KISS = '\U0001f48b'
    UNICODE_KISS_CHARACTER = '💋'
    UNICODE_KISS_SURROGATE_PAIR = '\ud83d\udc8b'


class EmojiUnicodeTest(UnicodeTestBase):
    def test_should_also_store_the_unicode_of_an_emoji(self):
        self.assertEqual(len(Emoji._unicode_characters), TOTAL_EMOJIS_UNICODE)

    def test_should_be_able_to_look_up_unicode(self):
        self.assertEqual(Emoji.name_for(self.UNICODE_KISS), 'kiss')
        self.assertEqual(Emoji.name_for(self.UNICODE_KISS_CHARACTER), 'kiss')

    def test_replace_unicode_in_string_with_images(self):
        self.assertIn(
            'kiss.png',
            Emoji.replace_unicode('I send a {0}!'.format(self.UNICODE_KISS)),
        )

    def test_emojis_that_were_showing_wrong(self):
        # These emojis were showing other images than the ones intended.
        emoji = '🍆'
        self.assertEqual(Emoji.name_for(emoji), 'eggplant')

        emoji = '😘'
        self.assertEqual(Emoji.name_for(emoji), 'kissing_heart')

    def test_remove_unicode_control_from_name_for(self):
        # fe0f is a control character that modifies the character
        # before, but we already do that modification so it's just
        # unecessary.
        emoji = '✌️'
        self.assertEqual(Emoji.name_for(emoji), 'v')

    def test_remove_unicode_control_from_replace_unicode(self):
        emoji = '✌️'
        self.assertEqual(
            Emoji.replace_unicode(emoji),
            '<img src="/static/emoji/img/v.png" alt="✌" '
            'title="v" class="emoji">'
        )

    def test_can_convert_unicode_surrogate_pair_emoji(self):
        """When in narrow mode the output characters might be a surrogate
        pair, but that matches in string comparison with the wide
        character.

        """
        # Tests that only matter when running narrow, @skipIf not
        # available in 2.6.
        if UNICODE_WIDE:
            return

        res = Emoji.replace_unicode(self.UNICODE_KISS_SURROGATE_PAIR)
        self.assertIn('\ud83d\udc8b', res)
        self.assertIn('\U0001f48b', res)


class EmojiHtmlEntitiesTest(UnicodeTestBase):
    INTEGER_KISS = '&#128139;'
    HEX_KISS = '&#x0001f48b;'
    HEX_KISS_SHORT = '&#x1f48b;'

    def test_should_replace_integer_encoded_unicode(self):
        self.assertEqual(
            Emoji.replace_html_entities(self.INTEGER_KISS),
            self.UNICODE_KISS
        )

    def test_should_replace_hex_encoded_unicode(self):
        self.assertEqual(
            Emoji.replace_html_entities(self.HEX_KISS),
            self.UNICODE_KISS
        )

    def test_should_replace_short_hex_encoded_unicode(self):
        self.assertEqual(
            Emoji.replace_html_entities(self.HEX_KISS_SHORT),
            self.UNICODE_KISS
        )
